//! Application-local prelude: conveniently import types/functions/macros
//! which are generally useful and should be available everywhere.

/// Abscissa core prelude
pub use abscissa_core::prelude::*;

/// Application state
pub use crate::application::APP;
